use std::collections;

use crate::node;
use crate::parser;

struct Context
{
    variable_hashmap : collections::HashMap<String, i64>
}

fn run_node(context : &mut Context, node : node::Node) -> i64
{
    match node
    {
        node::Node::Number(value) => value,

        node::Node::Expression(operator, left_value, right_value) =>
        {
            return calculate(operator, run_node(context, *left_value), run_node(context, *right_value));
        },

        node::Node::GetVariable(variable_name) =>
        {
            return match context.variable_hashmap.get(&variable_name)
            {
                Some(value) => *value,
                None => 0
            };
        },

        node::Node::SetVariable(variable_name, node) =>
        {
            let value : i64 = run_node(context, *node);

            context.variable_hashmap.insert(variable_name, value);

            return value;
        },

        node::Node::If(condition_node_box, true_node_vector_box, false_node_vector_box) =>
        {
            let result : i64 = run_node(context, *condition_node_box);

            if result > 0
            {
                return run_nodes(context, &*true_node_vector_box);
            }
            else
            {
                return run_nodes(context, &*false_node_vector_box);
            }
        },

        node::Node::For(control_variable_name, start_value, end_value, body_node_vector_box) =>
        {
            let mut result : i64 = 0;

            let node_vector : Vec<node::Node> = *body_node_vector_box;

            for i in start_value..=end_value
            {
                context.variable_hashmap.insert(control_variable_name.clone(), i);

                result = run_nodes(context, &node_vector);
            }

            return result;
        },

        node::Node::PrintString(value) =>
        {
            println!("{}", value);

            return 0;
        },

        node::Node::Print(node_box) =>
        {
            let value : i64 = run_node(context, *node_box);

            println!("{}", value);

            return value;
        },

        _ => 0
    }
}

fn calculate(operator : char, left_value : i64, right_value : i64) -> i64
{
    return match operator
    {
        '+' => left_value + right_value,
        '-' => left_value - right_value,
        '*' => left_value * right_value,
        '/' => left_value / right_value,
        '%' => left_value % right_value,
        '=' => if left_value == right_value {1} else {0},
        '!' => if left_value != right_value {1} else {0},
        '>' => if left_value >  right_value {1} else {0},
        'g' => if left_value >= right_value {1} else {0},
        '<' => if left_value <  right_value {1} else {0},
        'l' => if left_value <= right_value {1} else {0},
        _   => 0
    };
}

fn run_nodes(context : &mut Context, node_vector : &Vec<node::Node>) -> i64
{
    let mut result : i64 = 0;

    node_vector.iter().for_each(|node : &node::Node| { result = run_node(context, node.clone()); });

    return result;
}

pub fn run(source : &str) -> i64
{
    let node_vector : Vec<node::Node> = parser::custom_language::parse(source).unwrap();

    let mut context : Context = Context { variable_hashmap : collections::HashMap::new() };

    return run_nodes(&mut context, &node_vector);
}